<?php
/**
 *  Fluistr addon Class for the plugin Wordfence
 *
 *  @class WP_Fluistr_Addon_Itsec
 *  @package fluistr-athentication
 *  @subpackage addons
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}


/**
 * Loads load_mod_fluistr_addon_wordfence
 *
 * @return void
 */
function load_mod_fluistr_addon_wordfence() {
	if (
		empty( WPFluistrAuth::get_option( 'fluistr_activated' ) )
		|| class_exists( 'WP_Fluistr_Addon_Wordfence' )
		|| ! class_exists( 'wordfence' ) ) {
		return;
	}

	/**
	 * Class WP_Fluistr_Addon_Wordfence
	 */
	class WP_Fluistr_Addon_Wordfence {
		/**
		 * __construct
		 *
		 * @return void
		 */
		public function __construct() {

			WPFluistrAuth::AddModule(
				'fluistr_wordfence',
				array(
					'name'    => 'Fluistr Addon for Wordfence',
					'version' => '0.6.0',
				)
			);

			add_action( 'fluistr_validate_username_failed', array( $this, 'valide_username_failed' ), 10, 2 );
			add_action( 'fluistr_appsignon', array( $this, 'appsignon' ), 11, 1 ); // login via app
			add_action( 'fluistr_applogout', array( $this, 'applogout' ), 11, 1 ); // logout via app
		}

		/**
		 * Handles a validate_username_failed
		 *
		 * @param mixed $username WP username.
		 * @param mixed $error .
		 * @return void
		 */
		public function validate_username_failed( $username = '', $error = '' ) {
			try {
				wordfence::authenticateFilter( null, $username, '' );
			} catch ( Exception $e ) {
				if ( WP_DEBUG() ) {
					throw new \Exception( $e->getMessage() );
				}
			}
		}

		/**
		 * App signon
		 *
		 * @param mixed $user .
		 * @return void
		 */
		public function appsignon( $user = null ) {
			try {
				// needed to bypass default skipping.
				if ($_POST){
					$_POST['action'] = 'appsignon';
					wordfence::loginAction( $user );
					unset( $_POST['action'] );
				}
			} catch ( Exception $e ) {
				if ( WP_DEBUG() ) {
					throw new \Exception( $e->getMessage() );
				}
			}
		}

		/**
		 * App logout
		 *
		 * @param mixed $user_id WP user id.
		 * @return void
		 */
		public function applogout( $user_id = null ) {
			try {
				wordfence::logoutAction();
			} catch ( Exception $e ) {
				if ( WP_DEBUG() ) {
					throw new \Exception( $e->getMessage() );
				}
			}
		}
	}

	new WP_Fluistr_Addon_Wordfence();
}

add_action( 'wp_fluistr_modules', 'load_mod_fluistr_addon_wordfence' );
