<?php
/**
 *  Fluistr Authentication
 *
 *  @package fluistr-athentication
 *  @version 1.0.1
 *  @subpackage fluistr-login
 */

if ( ! class_exists( 'FluistrLoginController', false ) ) {
	/**
	 *  Fluistr Authentication
	 *
	 *  @class FluistrLoginController
	 *  @package fluistr-athentication
	 *  @version 1.0.1
	 *  @subpackage fluistr-login
	 */
	class FluistrLoginController {
		/**
		 * __construct
		 *
		 * @return void
		 */
		public function __construct() {
			if ( 'wp-login.php' === $GLOBALS['pagenow'] && ! empty( $_REQUEST['action'] ) && in_array( $_REQUEST['action'], array( 'register', 'lostpassword', 'resetpass' ), true ) ) {
				if ( WPFluistrAuth::stealth_modus() ) {
					wp_die( __( 'Password recovery disabled by Fluistr.', 'fluistr-auth' ) );
				}

				if ( ! isset( $_REQUEST['user_login'] ) ) {
					return;
				}
				$user = $this->get_user_from_sanitized_request( $_REQUEST['user_login'] );
				if ( empty( $user ) ) {
					return;
				}

				if ( empty( get_user_meta( $user->ID, 'fluistr_activated', true ) ) ) {
					return;
				}
				wp_die( __( 'Password recovery not allowed for this activated fluistr user account.', 'fluistr-auth' ) );
			}

			if ( 'wp-login.php' === $GLOBALS['pagenow'] && ! empty( $_REQUEST['action'] ) && 'rp' === $_REQUEST['action'] ) {

				if ( ! isset( $_REQUEST['login'] ) ) {
					return;
				}
				$user = $this->get_user_from_sanitized_request( $_REQUEST['login'] );
				if ( empty( $user ) ) {
					return;
				}

				if ( empty( get_user_meta( $user->ID, 'fluistr_activated', true ) ) ) {
					return;
				}
				if ( ! WPFluistrAuth::stealth_modus() ) {
					return;
				}
				auth_redirect();
				die;
			}

			add_action( 'wp_ajax_nopriv_fluistr_validate_username', array( $this, 'ajax_validate_username' ) );
			add_action( 'wp_ajax_fluistr_validate_username', array( $this, 'ajax_validate_username' ) );

			add_action( 'wp_ajax_fluistr_check', array( $this, 'ajax_check' ) );
			add_action( 'wp_ajax_nopriv_fluistr_check', array( $this, 'ajax_check' ) );

			add_action( 'wp_ajax_fluistr_disconnect', array( $this, 'ajax_disconnect' ) );
			add_action( 'wp_ajax_nopriv_fluistr_disconnect', array( $this, 'ajax_disconnect' ) );

			add_action( 'login_enqueue_scripts', array( $this, 'enqueue_scripts' ) );
			add_filter( 'login_headertext', array( $this, 'login_headertext' ), 10, 4 );
			add_filter( 'login_body_class', array( $this, 'login_body_class' ), 10, 4 );
			add_filter( 'login_body_class', array( $this, 'login_body_class' ), 10, 4 );

			add_action( 'login_header', array( $this, 'add_views' ), 10 );

			add_action( 'wp_logout', array( $this, 'wp_logout' ) );

			global $pagenow;
			if ( ! empty( $pagenow ) && ! empty( $_GET['action'] ) ) :

				$action = sanitize_key( $_GET['action'] );
				$urlnow = basename( self::fluistr_login_url() );
				if ( strpos( $urlnow, '?' ) !== false ) :
					$url_arr = explode( '?', $urlnow );
					$urlnow  = $url_arr[0];
				endif;

				if ( $urlnow == $pagenow ) :
					$this->switch_login_action( $action );
				endif;
			endif;
		}


		/**
		 * Return sanitizedd user from request
		 *
		 * @param  string $req_login name.
		 * @return mixed WP user or false.
		 */
		public function get_user_from_sanitized_request( $req_login = '' ) {
			$user_login = sanitize_user( $req_login );
			$user       = get_user_by( 'login', $user_login );
			if ( empty( $user ) ) {
				$user_email = $this->sanitize_email( $req_login );
				$user       = get_user_by( 'email', $user_email );
			}
			return $user;
		}

		/**
		 * Sanitizes email or returns an empty string
		 *
		 * @param  string $email string.
		 * @return string
		 */
		public function sanitize_email( $email = '' ) {
			if ( ! filter_var( $email, FILTER_VALIDATE_EMAIL ) ) {
				$email = '';
			}
			return $email;
		}


		/**
		 * Sanitizes door string
		 *
		 * @param  string $door string.
		 * @return string
		 */
		public function sanitize_door( $door = '' ) {
			$door = preg_replace( '/[^a-zA-Z0-9]/', '', $door );
			return $door;
		}

		/**
		 * Sanitizes door key string
		 *
		 * @param  string $key string.
		 * @return string
		 */
		public function sanitize_door_key( $key = '' ) {
			$key = sanitize_text_field( $key );
			return $key;
		}

		/**
		 * Returns the fluistr login url.
		 *
		 * @return string
		 */
		public static function fluistr_login_url() {
			$url = wp_login_url();
			if ( strpos( $url, '?' ) !== false ) :
				$url .= '&';
			else :
				$url .= '?';
			endif;
			return $url;
		}


		/**
		 * Returns login headertext.
		 *
		 * @return string
		 */
		public function login_headertext( $login_headertext = '' ) {
			return __( 'Fluistr Authentication', 'fluistr-auth' );
		}


		/**
		 * Returns extra classes for the login page.
		 *
		 * @param array  $classes
		 * @param string $action
		 * @return string
		 */
		public function login_body_class( $classes, $action = '' ) {
			$new_classes = array();
			foreach ( $classes as $c ) :
				if ( $c != 'login-action-login' ) {
					$new_classes[] = $c;
				}
			endforeach;
			return $new_classes;
		}


		public function enqueue_scripts() {

			wp_enqueue_style(
				'wpfluistr-login',
				plugins_url( '/modules/fluistr-login/assets/fluistr-login.css', WP_FLUISTR_PLUGIN_BASENAME ),
				null,
				WP_FLUISTR_VERSION
			);

			wp_enqueue_script(
				'qrcodejs-min', // Handle.
				plugins_url( '/modules/fluistr-login/assets/qrcode.min.js', WP_FLUISTR_PLUGIN_BASENAME ),
				null,
				WP_FLUISTR_VERSION, // Version: File modification time.
				true
			);

			wp_enqueue_script(
				'wpfluistr-login', // Handle.
				plugins_url( '/modules/fluistr-login/assets/fluistr-login.min.js', WP_FLUISTR_PLUGIN_BASENAME ),
				array( 'jquery' ), // Dependencies, defined above.
				WP_FLUISTR_VERSION, // Version: File modification time.
				true // Enqueue the script in the footer.
			);

			wp_localize_script(
				'wpfluistr-login',
				'fluistr',
				array(
					'ajaxurl' => admin_url( 'admin-ajax.php' ),
					'nonce'   => wp_create_nonce( 'validate_fluistrlogin' ),
					'expires' => intval( WPFluistrAuth::get_option( 'expires' ) ),
				)
			);
		}

		/**
		 * Ajax validation of username.
		 *
		 * @param string $nonce
		 * @param string $mail_type
		 * @param string $user_login
		 * @param string $rememberme
		 * @output json
		 */
		public function ajax_validate_username() {

			if ( ! isset( $_REQUEST['nonce'] )
			|| ! wp_verify_nonce( $_REQUEST['nonce'], 'validate_fluistrlogin' ) ) {
				wp_die( 'don\'t try to beat the nonce' );
			}

			$mail_type  = isset( $_REQUEST['mail_type'] ) && in_array( $_REQUEST['mail_type'], array( 'setup', 'magiclink' ), true ) ? $_REQUEST['mail_type'] : 'setup';
			$username   = isset( $_REQUEST['user_login'] ) ? sanitize_text_field( $_REQUEST['user_login'] ) : '';
			$rememberme = ! empty( $_REQUEST['rememberme'] ) ? true : false;
			$message    = '';

			$user = get_user_by( 'login', $username );

			if ( empty( $user ) ) {
				$user = get_user_by( 'email', $username );
			}

			if ( empty( $user ) ) {
				$error_message = __( '<strong>Error</strong>: Invalid username, email address or incorrect password.' );
				$error         = new WP_Error( 'authentication_failed', $error_message );

				do_action( 'fluistr_validate_username_failed', $username, $error );
				do_action( 'wp_login_failed', $username, $error );

				wp_send_json(
					(object) array(
						'fluistr' => false,
						'status'  => 404,
						'message' => $error_message,
					)
				);
			} else {
				update_metadata( 'user', $user->ID, 'rememberme', $rememberme );

				if ( 'magiclink' === $mail_type ) {
					$message = '<strong>' . __( 'An e-mail with the Magic Link is sent to your mail account.', 'fluistr-auth' ) . '</strong>';
					require_once WP_FLUISTR_PLUGIN_DIR . '/modules/fluistr-login/pluggable.php';

					$send = fluistr_retrieve_challenge( $user, $mail_type );

					if ( is_wp_error( $send ) ) {

						wp_send_json(
							(object) array(
								'fluistr' => false,
								'status'  => 404,
								'message' => $send->get_error_message(),
							)
						);
					}

					wp_send_json(
						(object) array(
							'fluistr' => true,
							'status'  => 221,
							'message' => $message,
						)
					);
				}
				$fluistr_activated = get_user_meta( $user->ID, 'fluistr_activated', true );

				if ( $fluistr_activated || WPFluistrAuth::stealth_modus() ) {
					$fluistr_door = get_user_meta( $user->ID, 'fluistr_door', true );

					// fluistr routing.
					if ( empty( $fluistr_door ) || ! empty( get_user_meta( $user->ID, 'fluistr_init' ) ) ) {
						// new
						$message = '<strong>' . __( 'Check your e-mail for further instructions on how to login with Fluistr.', 'fluistr-auth' ) . '</strong>';
						$result  = (object) array(
							'fluistr' => true,
							'status'  => 221,
							'message' => $message,
						);
						require_once WP_FLUISTR_PLUGIN_DIR . '/modules/fluistr-login/pluggable.php';

						$send = fluistr_retrieve_challenge( $user, $mail_type );

						if ( is_wp_error( $send ) ) {

							wp_send_json(
								(object) array(
									'fluistr' => false,
									'status'  => 404,
									'message' => $send->get_error_message(),
								)
							);
						}

						update_metadata( 'user', $user->ID, 'fluistr_init', time() );

					} else {
						// check via app.
						$fluistr_nonce = $this->open_user_door( $user );
						$result        = (object) array(
							'fluistr' => $fluistr_nonce,
							'status'  => 220,
							'message' => '<strong>' . __( 'You can now authenticate in the Fluistr app to login safely.', 'fluistr-auth' ) . '</strong>',
						);
					}
				} else {
					// default routing.
					$result = (object) array(
						'fluistr' => false,
						'status'  => 200,
						'message' => __( 'Valid username.', 'fluistr-auth' ),
					);
				}
			}
			wp_send_json( $result );
		}


		/**
		 * Ajax disconnect.
		 *
		 * @param string $nonce .
		 * @output json
		 */
		public function ajax_disconnect() {
			$connection_corrupted = __( 'Connection corrupted. Please connect again.', 'fluistr-auth' );

			$nonce   = sanitize_text_field( $_POST['nonce'] );
			$user_id = $this->get_user_id_from_nonce( $nonce );
			if ( ! $user_id && empty( $_POST['username'] ) ) {
				wp_send_json( array( 'error' => $connection_corrupted ) );
			}

			if ( ! $user_id ) :
				$user = $this->get_user_from_sanitized_request( $_POST['username'] );

				if ( ! $user || empty( $user->ID ) ) {
					wp_send_json( array( 'error' => $connection_corrupted ) );
				}
				$user_id = $user->ID;
			endif;

			$this->close_user_door( $user_id, true );
			delete_metadata( 'user', $user_id, 'fluistr_door' );

			wp_send_json(
				array(
					'error' => '',
					'url'   => $url,
					'nonce' => $nonce,
				)
			);
		}


		/**
		 * Timed ajax check.
		 *
		 * @output json
		 */
		public function ajax_check() {
			$connection_corrupted = __( 'Connection corrupted. Please connect again.', 'fluistr-auth' );
			if ( ! isset( $_POST['nonce'] ) ) {
				wp_send_json( array( 'error' => $connection_corrupted ) );
			}

			$nonce   = sanitize_text_field( $_POST['nonce'] );
			$user_id = $this->get_user_id_from_nonce( $nonce );

			if ( ! $user_id ) {
				wp_send_json( array( 'error' => $connection_corrupted ) );
			}

			$door_status = get_user_meta( $user_id, 'door_opened', true );
			if ( empty( $door_status ) ) {
				wp_send_json( array( 'error' => $connection_corrupted ) );
			}

			$url = '';
			if ( 'YES' === $door_status ) {
				$nonce = '';
				delete_metadata( 'user', $user_id, 'fluistr_init' );
				$enter_key = $this->set_user_key( 'user_activation_key', $user_id );
				$url       = self::fluistr_login_url() . 'action=enter&entry_key=' . rawurlencode( $enter_key );
			} else {
				$nonce = $this->set_user_key( 'user_activation_key', $user_id );
			}

			wp_send_json(
				array(
					'error' => '',
					'url'   => $url,
					'nonce' => $nonce,
				)
			);
		}


		/**
		 * This switch is for all action found in de url wp-login.php?action=...
		 *
		 * @param string $action
		 */
		private function switch_login_action( $action ) {

			$login        = isset( $_GET['login'] ) ? sanitize_text_field( $_GET['login'] ) : null;
			$key          = isset( $_GET['key'] ) ? $this->sanitize_door_key( $_GET['key'] ) : null;
			$entry_key    = isset( $_GET['entry_key'] ) ? sanitize_text_field( $_GET['entry_key'] ) : null;
			$post_door    = isset( $_POST['d'] ) ? $this->sanitize_door( $_POST['d'] ) : '';
			$post_email   = isset( $_POST['e'] ) ? $this->sanitize_email( $_POST['e'] ) : '';
			$post_message = isset( $_POST['m'] ) ? sanitize_textarea_field( $_POST['m'] ) : '';
			$post_n       = isset( $_GET['n'] ) ? sanitize_text_field( $_GET['n'] ) : '';

			switch ( $action ) :
				case 'activate':
					if ( ! empty( $login ) && ! empty( $key ) ) {
						$this->rp_activate( $key, $login );
					}

					$this->activate();

					break;
				case 'validating':
					if ( ! $this->validating( $key, $post_n ) ) {
						$this->has_error = __( 'invalid authentication', 'fluistr-auth' );
						do_action( 'wp_login_failed', $post_n, $this->has_error );
					}
					break;
				case 'authenticate':
					$this->authenticate( $post_door, $post_email, $post_message );
					break;
				case 'enter':
					if ( is_user_logged_in() ) {
						wp_safe_redirect( get_dashboard_url() );
						die();
					}
					$this->enter_user_door( $entry_key );
					break;
				case 'check':
					$this->check_loggedin_status();
					break;
				case 'logout':
					if ( ! empty( $post_door ) && ! empty( $post_email ) ) {
						$this->logout_via_app();
					}

					break;

				case 'rp':
					break;
				case 'lostpassword':
				case 'error':
					if ( isset( $_GET['e'] ) ) {
						WPFluistrAuth::AddNotice( '<strong>' . __( 'ERROR', 'fluistr-auth' ) . '</strong>: ' . __( 'Invalid activation link.', 'fluistr-auth' ), 'error', true );
						// put in login_footer for very late notices to show up.
						add_action( 'login_footer', array( 'WPFluistrAuth', 'admin_notices' ) );
					}

					// default password recovery actions. Do nothing for now.
					break;
				default:
					$this->has_error = __( 'Empty fluistr', 'fluistr-auth' );
			endswitch;
		}



		/**
		 * Sets the rp_cookie.
		 *
		 * @param string $rp_key
		 * @param string $rp_login
		 */

		public function rp_activate( $rp_key, $rp_login ) {

			list( $rp_path ) = explode( '?', wp_unslash( $_SERVER['REQUEST_URI'] ) );
			$rp_cookie       = 'wp-resetpass-' . COOKIEHASH;
			if ( isset( $rp_key ) ) {
				$value = sprintf( '%s:%s', wp_unslash( $rp_login ), wp_unslash( $rp_key ) );
				@setcookie( $rp_cookie, $value, 0, $rp_path, COOKIE_DOMAIN, is_ssl(), true );
				wp_safe_redirect( remove_query_arg( array( 'key', 'login' ) ) );
				exit;
			}
		}

		/**
		 * Start activation process.
		 */
		public function activate() {

			/**
			 * Stripped from wp-login.php.
				case 'resetpass' :
				case 'rp' :
			*/
			$user = false;

			list( $rp_path ) = explode( '?', wp_unslash( $_SERVER['REQUEST_URI'] ) );
			$rp_cookie       = 'wp-resetpass-' . COOKIEHASH;

			if ( isset( $_COOKIE[ $rp_cookie ] ) && 0 < strpos( $_COOKIE[ $rp_cookie ], ':' ) ) {
				list( $rp_login, $rp_key ) = explode( ':', wp_unslash( $_COOKIE[ $rp_cookie ] ), 2 );

				$user = @check_password_reset_key( $rp_key, $rp_login );
			}

			if ( ! $user || is_wp_error( $user ) ) {
				@setcookie( $rp_cookie, ' ', time() - YEAR_IN_SECONDS, $rp_path, COOKIE_DOMAIN, is_ssl(), true );

				wp_safe_redirect( self::fluistr_login_url() . 'action=error&e=expired_activate_link' );
				die();

			} else {

				if ( isset( $_GET['direct'] ) ) {
					update_metadata( 'user', $user->ID, 'logged_in', true );
					$secure_cookie = is_ssl();
					$rememberme    = get_metadata( 'user', $user->ID, 'rememberme', true ) ? true : false;

					// remove temp token.
					$this->set_user_key( 'user_activation_key', $user->ID, '' );

					wp_set_current_user( $user->ID, $user->user_login );
					@wp_set_auth_cookie( $user->ID, $rememberme, $secure_cookie );
					do_action( 'wp_login', $user->user_login, $user ); // second arg is needed for some plugins
					wp_safe_redirect( get_dashboard_url() );

				} else {

					global $fluistr_qrcode_uri;
					global $fluistr_nonce;

					// md5 only used here for formatting a nicer token string.
					$door = md5( wp_generate_password( 20, true, true ) );
					update_metadata( 'user', $user->ID, 'fluistr_door', $door );
					update_metadata( 'user', $user->ID, 'fluistr_init', time() );

					$message = (object) array(
						'email'       => $user->user_email,
						'sitename'    => get_bloginfo( 'name' ),
						'fluistr_uri' => wp_login_url(),
						'door'        => $door,
					);

					$fluistr_nonce = $this->open_user_door( $user );

					$this->set_user_key( 'user_pass', $user->ID, $fluistr_nonce );

					require_once WP_FLUISTR_PLUGIN_DIR . '/libs/cryptojs.php';
					try {
						$m = fluistrPHPEncrypt( $fluistr_nonce, wp_json_encode( $message ) );
					} catch ( Exception $e ) {
						die; }

					$fluistr_qrcode_uri = 'k=' . rawurlencode( $fluistr_nonce ) . '&m=' . urlencode( $m );
					add_action( 'login_header', array( $this, 'activate_add_to_loginfooter' ), 10 );
				}
			}
		}

		/**
		 * Adds the activation view.
		 */
		public function activate_add_to_loginfooter() {
			require_once WP_FLUISTR_PLUGIN_DIR . '/modules/fluistr-login/views/activate.php';
		}


		/**
		 * Adds the views components on the login page
		 */
		public function add_views() {
			require_once WP_FLUISTR_PLUGIN_DIR . '/modules/fluistr-login/views/login.php';
		}


		/**
		 * Gets the user id from the nonce.
		 *
		 * @param string $nonce
		 * @return false or int
		 */
		private function get_user_id_from_nonce( $nonce ) {
			global $wpdb;

			$res = $wpdb->get_row( $wpdb->prepare( "SELECT ID FROM {$wpdb->users} WHERE user_activation_key = %s", $nonce ) );
			if ( ! empty( $wpdb->error ) ) {
				// There was an error.
				wp_die( __( 'Error setting activation key', 'fluistr-auth' ) );
			} elseif ( empty( $res ) ) {
				return false;
			} else {
				return $res->ID;
			}
		}


		/**
		 * Creates and sets a new userkey.
		 *
		 * @param string $field
		 * @param int    $user_id
		 * @param mixed  $value
		 * @return $mixed.
		 */
		private function set_user_key( $field = false, $user_id = false, $value = false ) {
			if ( ! $field || ! $user_id ) {
				wp_die( __( 'error setting activation key', 'fluistr-auth' ) );
			}
			global $wpdb;

			/**
			 * Validate $field for correct use as wpdb->users table field.
			 */
			if ( ! in_array( $field, array( 'user_activation_key', 'user_pass' ), true ) ) {
				wp_die( __( 'error validating user key', 'fluistr-auth' ) );
			}

			if ( false === $value ) {
				$res           = true;
				$breakitafther = 0;
				while ( ! empty( $res ) && $breakitafther < 20 ) {
					$breakitafther++;
					// the key exists, try again!
					$value = 'fluistr:' . wp_generate_password( 20, false );

					$res = $wpdb->get_col( $wpdb->prepare( "SELECT ID FROM {$wpdb->users} WHERE " . $field . ' = %s', $value ) );
					if ( ! empty( $wpdb->error ) ) {
						// There was an error.
						wp_die( __( 'Error setting user key ', 'fluistr-auth' ) );
					}
				}
			}
			// good to go.
			$wpdb->update( $wpdb->users, array( $field => $value ), array( 'ID' => $user_id ), array( '%s' ), array( '%d' ) );

			if ( ! empty( $wpdb->error ) ) {
				// There was an error.
				wp_die( 'Error setting user key ' );
			} else {
				return $value;
			}
		}



		/**
		 * Checks of the user is still logged in.
		 *
		 * @output json
		 */
		public function check_loggedin_status() {

			$door  = $this->sanitize_door( $_POST['d'] );
			$email = $this->sanitize_email( $_POST['e'] );

			if ( empty( $door ) || empty( $email ) ) {
				wp_send_json(
					array(
						'code'       => 404,
						'error'      => __( 'no valid credentials', 'fluistr-auth' ),
						'connection' => false,
					)
				);
			}

			$user = get_user_by( 'email', $email );
			if ( ! $user || is_wp_error( $user ) ) {
				wp_send_json(
					array(
						'code'      => 401,
						'error'     => __( 'no valid user', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			$user->door = get_user_meta( $user->ID, 'fluistr_door', true );
			if ( $user->door != $door ) {
				wp_send_json(
					array(
						'code'      => 402,
						'error'     => __( 'sorry, wrong door', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			$door_status = get_user_meta( $user->ID, 'door_opened', true );
			if ( empty( $door_status ) || $door_status !== 'YES' ) {
				wp_send_json(
					array(
						'code'      => 405,
						'error'     => __( 'sorry, door closed', 'fluistr-auth' ),
						'logged_in' => ! empty(
							get_metadata(
								'user',
								$user->ID,
								'logged_in',
								$user->exists()
							)
						),
					)
				);
			}

			wp_send_json(
				array(
					'code'      => 200,
					'error'     => '',
					'logged_in' => ! empty(
						get_metadata(
							'user',
							$user->ID,
							'logged_in',
							$user->exists()
						)
					),
				)
			);
		}



		/**
		 * Log out the use activated by app logout.
		 *
		 * @output json
		 */
		public function logout_via_app() {
			if ( empty( $_POST['d'] ) || empty( $_POST['e'] ) ) {
				wp_send_json(
					array(
						'code'       => 404,
						'error'      => __( 'no valid credentials', 'fluistr-auth' ),
						'connection' => false,
					)
				);
			}
			$door  = $this->sanitize_door( $_POST['d'] );
			$email = $this->sanitize_email( $_POST['e'] );

			$user = get_user_by( 'email', $email );
			if ( ! $user || is_wp_error( $user ) ) {
				wp_send_json(
					array(
						'code'      => 401,
						'error'     => __( 'no valid user', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			$user->door = get_user_meta( $user->ID, 'fluistr_door', true );

			if ( $user->door != $door ) {
				wp_send_json(
					array(
						'code'      => 402,
						'error'     => __( 'sorry, wrong door', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			update_metadata( 'user', $user->ID, 'logged_in', false );
			do_action( 'fluistr_applogout', $user->ID );

			// get all sessions for user with ID $user_id
			$sessions = WP_Session_Tokens::get_instance( $user->ID );

			// we have got the sessions, destroy them all!
			$sessions->destroy_all();

			wp_send_json(
				array(
					'code'      => 200,
					'error'     => '',
					'logged_in' => ! empty(
						get_metadata(
							'user',
							$user->ID,
							'logged_in',
							true
						)
					),
				)
			);
		}



		/**
		 * Authenticates the door values from the mobile app.
		 *
		 * @param string $door .
		 * @param string $email .
		 * @param string $message .
		 * @output json
		 */
		public function authenticate( $door = '', $email = '', $message = '' ) {
			if ( empty( $door ) || empty( $email ) || empty( $message ) ) {
				wp_send_json(
					array(
						'code'       => 404,
						'error'      => __( 'no valid credentials', 'fluistr-auth' ),
						'connection' => false,
					)
				);
			}
			$encrypted_message = stripslashes( $message );

			$user = get_user_by( 'email', $email );
			if ( ! $user || is_wp_error( $user ) || empty( $user->user_pass ) ) {
				wp_send_json(
					array(
						'code'      => 401,
						'error'     => __( 'no valid user', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			$user->door         = get_user_meta( $user->ID, 'fluistr_door', true );
			$user->door_expires = get_user_meta( $user->ID, 'door_expires', true );
			if ( empty( $user->door_expires ) ) {
				$user->door_expires = time() + 120;
			}
			delete_metadata( 'user', $user->ID, 'fluistr_init' );

			if ( intval( $user->door_expires ) < time() - 120 ) {
				$this->close_user_door( $user->ID, ! empty( $user->user_activation_key ) );
				wp_send_json(
					array(
						'code'      => 405,
						'error'     => __( 'sorry, door closed', 'fluistr-auth' ),
						'logged_in' => ! empty(
							get_metadata(
								'user',
								$user->ID,
								'logged_in',
								true
							)
						),
					)
				);
			}

			if ( $user->door != $door ) {
				wp_send_json(
					array(
						'code'      => 402,
						'error'     => __( 'sorry, wrong door', 'fluistr-auth' ),
						'logged_in' => false,
					)
				);
			}

			// can be YES or NO otherwise fail.
			$door_status = get_user_meta( $user->ID, 'door_opened', true );
			if ( empty( $door_status ) ) {
				wp_send_json(
					array(
						'code'      => 405,
						'error'     => __( 'sorry, door closed', 'fluistr-auth' ),
						'logged_in' => ! empty(
							get_metadata(
								'user',
								$user->ID,
								'logged_in',
								$user->exists()
							)
						),
					)
				);
			}

			/* DECRYPT RECEIVED INFO */

			require_once WP_FLUISTR_PLUGIN_DIR . '/libs/cryptojs.php';

			try {
				$message_txt = fluistrPHPDecrypt( $user->user_pass, $encrypted_message );

				$message = (array) json_decode( $message_txt, true );
			} catch ( Exception $e ) {
				wp_die( 'Oops, Decrypting message failed.' );
			}

			if ( empty( $message['new_passphrase'] ) || empty( $message['passphrase'] ) || $message['passphrase'] != $user->user_pass ) {
				// bad password.
				$this->close_user_door( $user->ID, ! empty( $user->user_activation_key ) );
				wp_send_json(
					array(
						'code'        => 403,
						'error'       => __( 'no valid doorkey', 'fluistr-auth' ),
						'logged_in'   => false,
						'p'           => $user->user_pass,
						'message_bad' => wp_json_encode( $message ),
					)
				);

			} elseif ( $this->update_fluistr( $user, $message['new_passphrase'] ) ) {
				// YES! you're in.
				update_metadata( 'user', $user->ID, 'door_opened', 'YES' );

				wp_send_json(
					array(
						'code'  => 200,
						'error' => '',
					)
				);
			}

			// Let's stop here. don't know what went wrong :-/.
			$this->close_user_door( $user->ID, true );
			wp_send_json(
				array(
					'code'      => 404,
					'error'     => __( 'no valid credentials.', 'fluistr-auth' ),
					'logged_in' => false,
				)
			);
		}


		/**
		 * Validate the doorkey. Get key vu
		 *
		 * @param string $key
		 * @return bool
		 */
		public function validating( $key = null, $nonce = '' ) {
			if ( empty( $key ) ) {
				return false;
			}

			$user_id = $this->get_user_id_from_nonce( $nonce );
			if ( ! $user_id ) {
				return false;
			}

			$user = get_user_by( 'id', $user_id );

			$this->fluistr_nonce = $this->open_user_door( $user );
			return true;
		}


		/**
		 * Creates the doorkey and activates the door.
		 *
		 * @param object $user
		 * @return string
		 */
		public function open_user_door( $user = null ) {
			if ( ! $user || is_wp_error( $user ) ) {
				return false;
			}
			$add_time = intval( WPFluistrAuth::get_option( 'expires' ) );
			$expires  = time() + ( ! empty( $add_time ) ? $add_time : 120 );
			$nonce    = $this->set_user_key( 'user_activation_key', $user->ID );

			update_metadata( 'user', $user->ID, 'door_expires', $expires );
			update_metadata( 'user', $user->ID, 'door_opened', 'NO' );
			return $nonce;
		}


		/**
		 * Updates the fluistr pass/key.
		 *
		 * @param object $user
		 * @param string $new_fluistr_pass
		 * @return bool
		 */
		public function update_fluistr( $user = null, $new_fluistr_pass = '' ) {
			// stop post because it will be used in signon function
			if ( isset( $_POST['user_login'] ) || isset( $_POST['user_password'] ) || ! $user || is_wp_error( $user ) || empty( $new_fluistr_pass ) ) {
				return false;
			}

			// set new pass.

			if ( strlen( $new_fluistr_pass ) < 20 ) {
				// password is invalid so for security reasons make a new one at random.
				$new_fluistr_pass = wp_generate_password( 20 );
				$new_fluistr_pass = 'fluistr:' . $new_fluistr_pass;

			}
			$this->set_user_key( 'user_pass', $user->ID, $new_fluistr_pass );

			return true;
		}


		/**
		 * Closes user_door
		 *
		 * @param  mixed $user_id
		 * @param  mixed $clear_activation_key
		 * @return void
		 */
		public function close_user_door( $user_id = false, $clear_activation_key = false ) {
			if ( ! $user_id ) {
				return false;
			}
			// close door.
			delete_metadata( 'user', $user_id, 'door_expires' );
			delete_metadata( 'user', $user_id, 'door_opened' );

			if ( $clear_activation_key ) {
				$this->set_user_key( 'user_activation_key', $user_id, '' );
			}
		}

		/**
		 * Validates doorkey and logs user in.
		 *
		 * @param  mixed $entry_key
		 * @return void
		 */
		function enter_user_door( $entry_key ) {
			if ( empty( $entry_key ) ) {
				wp_die( __( 'key not valid', 'fluistr-auth' ) );
			}

			$user_id = $this->get_user_id_from_nonce( $entry_key );
			if ( ! $user_id ) {
				wp_die( 'user not valid' );
			}

			$door_status = get_user_meta( $user_id, 'door_opened', true );

			if ( empty( $door_status ) || $door_status != 'YES' ) {
				wp_die( __( 'door not opened', 'fluistr-auth' ) );
			}

			$this->close_user_door( $user_id, true );

			$user = get_user_by( 'id', $user_id );

			if ( is_wp_error( $user ) ) {
				wp_die( __( 'error retrieving user' ) );
			}
			if ( $user ) {
				update_metadata( 'user', $user_id, 'logged_in', true );
				wp_set_current_user( $user_id, $user->user_login );

				do_action( 'fluistr_appsignon', $user );

				$secure_cookie  = is_ssl();
				$rememberme     = get_metadata( 'user', $user_id, 'rememberme', true ) ? true : false;
				$credentials    = array(
					'user_login'    => $user->user_login,
					'user_password' => $user->user_pass,
					'remember'      => $rememberme,
				);
				 $secure_cookie = apply_filters( 'secure_signon_cookie', $secure_cookie, $credentials );
				@wp_set_auth_cookie( $user->ID, $rememberme, $secure_cookie );
				/*
				 do not use:
				 * wp_signon()
				 * do_action( 'wp_login', $user->user_login,$user );
				 */
				wp_redirect( get_dashboard_url() );
			}
		}


		/**
		 * clears users fluistr metadata
		 */
		public function wp_logout( $user_id ) {

			/** custom signout */
			if ( ! empty( $user_id ) ) {
				delete_metadata( 'user', $user_id, 'logged_in' );
				delete_metadata( 'user', $user_id, 'door_expires' );
				delete_metadata( 'user', $user_id, 'door_opened' );
			}
		}
	}
}
